#!/usr/bin/env python

"""
Data representation functions.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from array import array
import operator

# High-level representations.

def convert_sequence(values, op, last_from_old):
    if values:
        new_values = list(values)
        last = new_values[0]
        i = 1
        length = len(new_values)
        while i < length:
            current = new_values[i]
            new_values[i] = op(current, last)

            # Subtracting entries requires the old value to be used.
            # Adding entries requires the new value.

            if last_from_old:
                last = current
            else:
                last = new_values[i]

            i += 1

        return new_values
    else:
        return values

def op_seq_monotonic(x, y, op):
    return tuple([op(a, b) for a, b in zip(x, y)])

def add_seq_monotonic(x, y):
    return op_seq_monotonic(x, y, operator.add)

def sub_seq_monotonic(x, y):
    return op_seq_monotonic(x, y, operator.sub)

def add_seq(x, y):
    length = min(len(x), len(y))
    seq = list(x)[:length]
    i = 0
    while i < length:
        if x[i] != 0:
            seq[i] = x[i] + y[i]
            break
        seq[i] = y[i]
        i += 1
    return tuple(seq)

def sub_seq(x, y):
    length = min(len(x), len(y))
    seq = list(x)[:length]
    i = 0
    while i < length:
        replacement = x[i] - y[i]
        if replacement != 0:
            seq[i] = replacement
            break
        seq[i] = 0
        i += 1
    return tuple(seq)

def is_sequence(value):
    return isinstance(value, (list, tuple))

def sizeof(value):
    return is_sequence(value) and len(value) or 0

def get_monotonic_adder(size):
    return size and add_seq_monotonic or operator.add

def get_monotonic_subtractor(size):
    return size and sub_seq_monotonic or operator.sub

def get_adder(size):
    return size and add_seq or operator.add

def get_subtractor(size):
    return size and sub_seq or operator.sub

# Low-level representations.
# Variable-length integer functions.

vint_cache = []
vint_bytes_cache = []

def vint(number):

    "Write 'number' as a variable-length integer."

    try:
        return vint_cache[number]
    except IndexError:
        if number >= 0:
            bytes = array('B')
            _vint_to_array(number, bytes)
            return bytes.tostring()

        # Negative numbers are not supported.

        else:
            raise ValueError, "Number %r is negative." % number

def vint_to_array(number, bytes):

    "Write 'number' as a variable-length integer to 'bytes'."

    try:
        bytes += vint_bytes_cache[number]
    except IndexError:
        if number >= 0:
            _vint_to_array(number, bytes)

        # Negative numbers are not supported.

        else:
            raise ValueError, "Number %r is negative." % number

def _vint_to_array(number, bytes):

    "Write the 'number' to 'bytes' from least to most significant digits."

    while number > 127:
        bytes.append(number & 127 | 128)
        number = number >> 7
    else:
        bytes.append(number)

def vint_from_array(bytes):

    "Read a variable-length integer from 'bytes', returning a number."

    number = 0
    while bytes:
        number <<= 7
        number += bytes.pop() & 127
    return number

def vint_from_array_start(bytes, start):

    """
    Read a variable-length integer from 'bytes', starting at 'start', and
    returning a tuple containing a number and the first position after the
    number.
    """

    length = len(bytes)
    if start == length:
        raise EOFError

    number = 0
    digit = 0
    while start < length:
        x = bytes[start]
        number += (x & 127) << digit
        digit += 7
        start += 1
        if not (x & 128):
            break
    return number, start

# Sequence serialisation.

def sequence_to_array(value, size, bytes):

    "Write the given sequence 'value' with the given 'size' to 'bytes'."

    if size:
        i = 0
        limit = min(len(value), size)
        while i < limit:
            vint_to_array(value[i], bytes)
            i += 1
        while i < size:
            vint_to_array(0, bytes)
    else:
        vint_to_array(value, bytes)

def sequence_from_array(bytes, size, start=0):

    """
    Read a sequence from 'bytes' having the given 'size', returning the sequence
    and the first position after the sequence.
    """

    if size:
        j = 0
        value = []
        while j < size:
            v, start = vint_from_array_start(bytes, start)
            value.append(v)
            j += 1
        return tuple(value), start
    else:
        return vint_from_array_start(bytes, start)

# Variable-length integer cache initialisation.

for i in xrange(0, 65536):
    bytes = array('B')
    _vint_to_array(i, bytes)
    vint_bytes_cache.append(bytes)
    vint_cache.append(bytes.tostring())

# vim: tabstop=4 expandtab shiftwidth=4
