#!/usr/bin/env python

"""
Generic file access.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.data import *
from array import array
import zlib

# Classes.

class File:

    "A basic file abstraction."

    def __init__(self, f):
        self.f = f
        self.record = array('B')    # record buffer
        self.data_start = None

    def begin(self):

        """
        Initialise file-wide parameters. In writers, this method may require
        parameters to be specified. In readers, the parameters may be read from
        the file.
        """

        self.data_start = 0

    def tell(self):
        # NOTE: Will not be accurate within the current record.
        return self.f.tell()

    def seek(self, offset):
        self.f.seek(offset)

    def rewind(self):
        self.seek(self.data_start)

    def close(self):
        if self.f is not None:
            self.f.close()
            self.f = None

class FileWriter(File):

    "Writing basic data types to files."

    def begin_record(self):
        pass

    def end_record(self):
        if self.record:
            self.f.write(vint(len(self.record)))
            self.record.tofile(self.f)
            self.record = array('B')

    def write_remaining(self, a):

        "Write remaining data from the raw array 'a'."

        self.record += a

    def write_byte(self, b):

        "Write the given byte 'b'."

        self.record.append(b)

    def write_number(self, number):

        "Write 'number' to the file using a variable length encoding."

        vint_to_array(number, self.record)

    def write_numbers(self, numbers):

        "Write 'numbers' to the file using a variable length encoding."

        for number in numbers:
            vint_to_array(number, self.record)

    def write_string(self, s, compress=0):

        """
        Write 's' to the file, recording its length and compressing the string
        if 'compress' is set to a true value.
        """

        # Convert Unicode objects to strings.

        if isinstance(s, unicode):
            s = s.encode("utf-8")

        # Compress the string if requested.

        if compress:
            cs = zlib.compress(s)

            # Take any shorter than the original.

            if len(cs) < len(s):
                flag = "z"
                s = cs
            else:
                flag = "-"

        else:
            flag = ""

        # Write the length of the data before the data itself.

        length = len(s)
        self.record.fromstring("".join([flag, vint(length), s]))

    def write_sequence_value(self, value, size):
        sequence_to_array(value, size, self.record)

    def write_sequence_values(self, values, size):
        vint_to_array(len(values), self.record)
        for value in values:
            self.write_sequence_value(value, size)

    def write_delta_sequence(self, values, size):
        self.write_sequence_values(
            convert_sequence(values, get_subtractor(size), 1),
            size)

    def write_monotonic_sequence(self, values, size):
        self.write_sequence_values(
            convert_sequence(values, get_monotonic_subtractor(size), 1),
            size)

    def close(self):
        self.end_record()
        File.close(self)

class FileReader(File):

    "Reading basic data types from files."

    def __init__(self, f):
        File.__init__(self, f)
        self.begin()

    def begin_record(self):
        self.start = 0
        self.record = array('B')
        try:
            size = self.read_number_from_file()
            self.record.fromfile(self.f, size)
        except EOFError:
            pass

    def end_record(self):
        pass

    def read_remaining(self):

        "Read remaining data as a raw array."

        return self.record[self.start:]

    def read_byte(self):

        "Read a byte from the record."

        b = self.record[self.start]
        self.start += 1
        return b

    def read_number_from_file(self):

        "Read a number from the file."

        # Read each byte, adding it to the number.

        a = array('B')
        a.fromfile(self.f, 1)
        csd = a[-1]
        if csd < 128:
            return csd
        else:
            while csd & 128:
                a.fromfile(self.f, 1)
                csd = a[-1]
            return vint_from_array(a)

    def read_number(self):

        "Read a number from the current record."

        n, self.start = vint_from_array_start(self.record, self.start)
        return n

    def read_numbers(self, n):
        l = []
        i = 0
        while i < n:
            l.append(self.read_number())
            i += 1
        return l

    def read_string(self, decompress=0):

        """
        Read a string from the current record, decompressing the stored data if
        'decompress' is set to a true value.
        """

        # Decompress the data if requested.

        if decompress:
            flag = chr(self.record[self.start])
            self.start += 1
        else:
            flag = "-"

        length = self.read_number()
        start = self.start
        self.start += length
        s = self.record[start:self.start].tostring()

        # Perform decompression if applicable.

        if flag == "z":
            s = zlib.decompress(s)

        # Convert strings to Unicode objects.

        return unicode(s, "utf-8")

    def read_sequence_value(self, size):
        value, self.start = sequence_from_array(self.record, size, self.start)
        return value

    def read_sequences(self, size):
        values = []
        length = self.read_number()
        i = 0
        while i < length:
            values.append(self.read_sequence_value(size))
            i += 1
        return values

    def read_delta_sequence(self, size):
        return convert_sequence(self.read_sequences(size), get_adder(size), 0)

    def read_monotonic_sequence(self, size):
        return convert_sequence(self.read_sequences(size), get_monotonic_adder(size), 0)

# vim: tabstop=4 expandtab shiftwidth=4
