#!/usr/bin/env python

"""
High-level classes.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.filesystem import *
from iixr.terms import MultipleReader
from itermerge import itermerge
from os import mkdir    # index discovery
from os.path import exists

# Constants.

FLUSH_INTERVAL    = 10000
INDEX_INTERVAL    = 100
OPEN_PARTITIONS   = 20

# High-level classes.

class Document:

    "A container of document information."

    def __init__(self, docnum):
        self.docnum = docnum
        self.terms = {}

    def add_position(self, term, position):

        """
        Add a position entry for the given 'term', indicating the given
        'position'.
        """

        self.terms.setdefault(term, []).append(position)

class IndexWriter:

    "Building term information and writing it to the term dictionary."

    def __init__(self, pathname, flush_interval):
        self.pathname = pathname
        self.flush_interval = flush_interval

        self.term_partition = get_next_partition(get_term_partitions(self.pathname))

        self.terms = {}
        self.doc_counter = 0

    def add_document(self, doc):

        """
        Add the given document 'doc', updating the document counter and flushing
        terms and fields if appropriate.
        """

        docnum = doc.docnum

        for term, positions in doc.terms.items():
            self.terms.setdefault(term, {})[docnum] = positions

        self.doc_counter += 1

        if self.flush_interval and self.doc_counter >= self.flush_interval:
            self.flush_terms()
            self.doc_counter = 0

    def get_term_writer(self):

        "Return a term writer for the current partition."

        return get_term_writer(self.pathname, self.term_partition)

    def flush_terms(self):

        "Flush terms into the current term partition."

        # Get the terms in order.

        term_writer = self.get_term_writer()
        try:
            term_writer.write_terms(self.terms)
        finally:
            term_writer.close()

        self.terms = {}
        self.term_partition += 1

    def close(self):
        if self.terms or not get_term_partitions(self.pathname):
            self.flush_terms()

class Index:

    "An inverted index solution encapsulating the various components."

    def __init__(self, pathname, flush_interval=FLUSH_INTERVAL,
        open_partitions=OPEN_PARTITIONS):

        self.pathname = pathname
        self.flush_interval = flush_interval
        self.open_partitions = open_partitions
        self.reader = None
        self.writer = None

    def get_writer(self):

        "Return a writer."

        if self.writer is None:
            self._ensure_directory()
            self.writer = IndexWriter(self.pathname, self.flush_interval)
        return self.writer

    def _ensure_directory(self):
        if not exists(self.pathname):
            mkdir(self.pathname)

    def _get_readers(self, get_reader):

        "Return a list of underlying readers given the 'get_reader' function."

        # Get the partitions in order.

        partitions = list(get_term_partitions(self.pathname))
        partitions.sort()

        # Return the readers.

        return [get_reader(self.pathname, partition) for partition in partitions]

    def get_reader(self, refresh=0):

        "Return a reader for the index."

        if refresh and self.reader is not None:
            self.reader.close()
            self.reader = None

        if self.reader is None:
            if not exists(self.pathname):
                raise OSError, "Index path %r does not exist." % self.pathname

            # Try and get combined readers.

            try:
                readers = self._get_readers(get_combined_term_reader)
            except IOError:
                readers = self._get_readers(get_term_reader)

            # NOTE: Multiple reading of combined readers is not supported.

            if len(readers) == 1:
                self.reader = readers[0]
            else:
                self.reader = MultipleReader(readers)

        return self.reader

    def merge(self, interval=INDEX_INTERVAL):

        "Merge the partitions in the index."

        # Get data readers.

        readers = self._get_readers(get_term_data_reader)
        reader = MultipleReader(readers, self.merge_data)
        writer = get_term_writer(self.pathname, "merged")
        index_writer = get_term_index_writer(self.pathname, "merged")

        try:
            writer.begin(*reader.get_sizes())
            index_writer.begin()

            i = 0
            for term, data in reader:
                if i % interval == 0:
                    index_writer.write_term(term, writer.tell())
                    index_writer.end_record()

                writer.write_term_plus_remaining(term, data)
                writer.end_record()
                i += 1

        finally:
            index_writer.close()
            writer.close()
            reader.close()

        for partition in get_term_partitions(self.pathname):
            remove_term_files(self.pathname, partition)

        rename_term_files(self.pathname, "merged", 0)
        self.reader = None
        self.writer = None

    def merge_data(self, a, b):

        """
        Merge 'a' and 'b', modifying the data to permit concatenation.
        """

        # Modify the record to indicate a continuation of the data.

        c = a + b
        c[len(a) - 1] = 1
        return c

    def close(self):
        if self.reader is not None:
            self.reader.close()
            self.reader = None
        if self.writer is not None:
            self.writer.close()
            self.writer = None

# vim: tabstop=4 expandtab shiftwidth=4
